# IMPORT LIBRARIES
import tkinter as TK
from .N_GUI import GUI
from .N_Frame import Frame
from .N_Custom import Event_Bind

class Text:

    def __init__(self, Main, Scroll=False, **kwargs):
        self._GUI = GUI._Instance
        if self._GUI is not None:
            self._Type = "Text"
            try:
                self._Config = ['Name', 'Auto_Dark', 'Background', 'Light_Background', 'Dark_Background', 'Foreground', 'Light_Foreground', 'Dark_Foreground', 'Border_Color', 'Light_Border_Color', 'Dark_Border_Color', 'Border_Size', 'Resize', 'Popup', 'Display', 'Left', 'Top', 'Width', 'Height', 'Font_Size', 'Font_Weight', 'Disable', 'Scrollbar_Active', 'Scrollbar_Size', 'Warp', 'Hover_Background', 'Light_Hover_Background', 'Dark_Hover_Background', 'Hover_Foreground', 'Light_Hover_Foreground', 'Dark_Hover_Foreground', 'Hover_Border_Color', 'Light_Hover_Border_Color', 'Dark_Hover_Border_Color']
                self._Initialized = False
                self._Name = False
                self._Last_Name = False
                self._Widget = []
                self._Tag = []
                self._Resize = True
                self._Popup = False
                self._Display = True
                self._Size_Update = False
                self._Main = Main
                self._Frame = Frame(self._Main)
                self._Widget = TK.Text(self._Frame._Frame)
                self._Scrollbar_Vertical = TK.Scrollbar(self._Frame._Frame)
                self._Scrollbar_Horizontal = TK.Scrollbar(self._Frame._Frame)
                self._Scrollbar_Size = 20
                self._Scrollbar_Active = False
                self._Warp = 'None'
                self._Border_Color = '#000000'
                self._Border_Size = 0
                self._Background = self._Main._Background
                self._Background_Main = True
                self._Foreground = '#000000'
                self._Hover_Background = False
                self._Hover_Foreground = False
                self._Hover_Border_Color = False
                self._Last_Background = False
                self._Last_Foreground = False
                self._Last_Border_Color = False
                self._Font_Size = 12
                self._Font_Weight = 'normal'
                self._Disable = False
                self._Auto_Dark = True
                self._On_Show = False
                self._On_Hide = False
            except Exception as E:
                self._GUI.Error(f"{self._Type} -> Init -> {E}")
        else:
            print("Error: Gluonix -> GUI Instance Has Not Been Created")

    def __str__(self):
        return "Nucleon_Glunoix_Text[]"

    def __repr__(self):
        return "Nucleon_Glunoix_Text[]"
    
    def Copy(self, Name=False, Main=False):
        try:
            if not Main:
                Main = self._Main
            Instance = type(self)(Main)
            for Key in self._Config:
                if hasattr(self, "_"+Key):
                    setattr(Instance, "_"+Key, getattr(self, "_"+Key))
            if Name:
                setattr(Instance, "_Name", Name)
            Instance.Create()
            return Instance
        except Exception as E:
            self._GUI.Error(f"{self._Type} -> Copy -> {E}")
        
    def Delete(self):
        try:
            self._Main._Widget.remove(self)
            self._Widget.destroy()
            self._Scrollbar_Vertical.destroy()
            self._Scrollbar_Horizontal.destroy()
            self._Frame.Delete()
            if self:
                del self
        except Exception as E:
            self._GUI.Error(f"{self._Type} -> Delete -> {E}")
            
    def Hide(self):
        try:
            self._Frame.Hide()
            self._Display = False
            if self._On_Hide:
                self._On_Hide()
        except Exception as E:
            self._GUI.Error(f"{self._Type} -> Hide -> {E}")
            
    def Show(self):
        try:
            self._Frame.Show()
            if self._Scrollbar_Active:
                self._Scrollbar_Vertical.place(relx=(self._Width-self._Scrollbar_Size)/self._Width, rely=0, relwidth=self._Scrollbar_Size/self._Width, relheight=1)
                if self._Warp.lower()=='none':
                    self._Scrollbar_Horizontal.place(relx=0, rely=(self._Height-self._Scrollbar_Size)/self._Height, relwidth=1, relheight=self._Scrollbar_Size/self._Height)
                    self._Widget.place(relx=0, rely=0, relwidth=(self._Width-self._Scrollbar_Size)/self._Width, relheight=(self._Height-self._Scrollbar_Size)/self._Height)
                else:
                    self._Scrollbar_Horizontal.place_forget()
                    self._Widget.place(relx=0, rely=0, relwidth=(self._Width-self._Scrollbar_Size)/self._Width, relheight=1)
            else:
                self._Scrollbar_Vertical.place_forget()
                self._Scrollbar_Horizontal.place_forget()
                self._Widget.place(relx=0, rely=0, relwidth=1.0, relheight=1.0)
            self._Display = True
            if self._On_Show:
                self._On_Show()
        except Exception as E:
            self._GUI.Error(f"{self._Type} -> Show -> {E}")
    
    def Grab(self, Path=False):
        try:
            return self._GUI.Grab_Widget(Path=Path, Widget=self)
        except Exception as E:
            self._GUI.Error(f"{self._Type} -> Grab -> {E}")
            
    def Animate(self, Hide=False, Thread=True):
        try:
            self._Frame.Animate(Widget=self._Widget, Thread=Thread)
        except Exception as E:
            self._GUI.Error(f"{self._Type} -> Animate -> {E}")
            self.Animate_Cancel()
            
    def Animate_Cancel(self):
        try:
            self._Frame.Animate_Cancel()
        except Exception as E:
            self._GUI.Error(f"{self._Type} -> Animate_Cancel -> {E}")
            
    def Get(self):
        try:
            return self._Widget.get("1.0", "end-1c")
        except Exception as E:
            self._GUI.Error(f"{self._Type} -> Get -> {E}")
            
    def Set(self, Value):
        try:
            self._Widget.config(state=TK.NORMAL)
            self._Widget.delete("1.0", "end")
            self._Widget.insert("1.0", Value)
            if self._Disable:
                State = TK.DISABLED
            else:
                State = TK.NORMAL
            self._Widget.config(state=State)
        except Exception as E:
            self._GUI.Error(f"{self._Type} -> Set -> {E}")
            
    def Add(self, Value, Tag='Default'):
        try:
            self._Widget.config(state=TK.NORMAL)
            self._Widget.insert("end", Value, Tag)
            if self._Disable:
                State = TK.DISABLED
            else:
                State = TK.NORMAL
            self._Widget.config(state=State)
        except Exception as E:
            self._GUI.Error(f"{self._Type} -> Add -> {E}")
            
    def Tag(self, Name, Font_Size=12, Font_Weight='normal', Font='Times New Roman', Foreground="black", Background="white"):
        try:
            if Name not in self._Tag:
                self._Widget.tag_config(Name, font=(Font, Font_Size, Font_Weight), foreground=Foreground, background=Background)
                self._Tag.append(Name)
        except Exception as E:
            self._GUI.Error(f"{self._Type} -> Tag -> {E}")
            
    def Widget(self):
        try:
            return self._Widget
        except Exception as E:
            self._GUI.Error(f"{self._Type} -> Widget -> {E}")
            
    def Bind(self, **Input):
        try:
            if 'On_Show' in Input:
                self._On_Show = Input['On_Show']
            if 'On_Hide' in Input:
                self._On_Hide = Input['On_Hide']
            Event_Bind(self._Widget, **Input)
        except Exception as E:
            self._GUI.Error(f"{self._Type} -> Bind -> {E}")
            
    def On_Hover_In(self, E):
        try:
            Config = {}
            if self._Hover_Background:
                self._Last_Background = self._Background
                Config['Background'] = self._Hover_Background
            if self._Hover_Foreground:
                self._Last_Foreground = self._Foreground
                Config['Foreground'] = self._Hover_Foreground
            if self._Hover_Border_Color:
                self._Last_Border_Color = self._Border_Color
                Config['Border_Color'] = self._Hover_Border_Color
            if len(Config)>0:
                self.Config(**Config)
        except Exception as E:
            self._GUI.Error(f"{self._Type} -> On_Hover_In -> {E}")
            
    def On_Hover_Out(self, E):
        try:
            Config = {}
            if self._Hover_Background and self._Last_Background:
                Config['Background'] = self._Last_Background if self._Background==self._Hover_Background else self._Background
            if self._Hover_Foreground and self._Last_Foreground:
                Config['Foreground'] = self._Last_Foreground if self._Foreground==self._Hover_Foreground else self._Foreground
            if self._Hover_Border_Color and self._Last_Border_Color:
                Config['Border_Color'] = self._Last_Border_Color if self._Border_Color==self._Hover_Border_Color else self._Border_Color
            if len(Config)>0:
                self.Config(**Config)
        except Exception as E:
            self._GUI.Error(f"{self._Type} -> On_Hover_Out -> {E}")
            
    def Config_Get(self, *Input):
        try:
            Return = {}
            for Each in self._Config:
                if Each in Input:
                    Return[Each] = getattr(self, "_"+Each)
            return Return
        except Exception as E:
            self._GUI.Error(f"{self._Type} -> Config_Get -> {E}")
                
    def Config(self, **Input):
        try:
            Run = False
            for Each in self._Config:
                if Each in Input:
                    Value = Input[Each]
                    setattr(self, "_"+Each, Value)
                    Run = True
            self._Frame.Config(**Input)
            if "Width" in Input or "Height" in Input or "Left" in Input or "Top" in Input:
                self._Size_Update = True
            if self._Initialized and Run:
                self.Create()
            if "Background" in Input:
                self._Background_Main = not bool(Input["Background"])
        except Exception as E:
            self._GUI.Error(f"{self._Type} -> Config -> {E}")
            
    def Move(self, Left=None, Top=None):
        try:
            if Left is not None:
                self._Left += Left
            if Top is not None:
                self._Top += Top
            if Left is not None or Top is not None:
                self.Position(Left=self._Left, Top=self._Top)
            return True
        except Exception as E:
            self._GUI.Error(f"{self._Type} -> Move -> {E}")
            
    def Center(self, Left=None, Top=None):
        try:
            if Left is not None:
                self._Left = Left-self._Width/2
            if Top is not None:
                self._Top = Top-self._Height/2
            if Left is not None or Top is not None:
                self.Position(Left=self._Left, Top=self._Top)
            return [self._Left+self._Width/2, self._Top+self._Height/2]
        except Exception as E:
            self._GUI.Error(f"{self._Type} -> Center -> {E}")
            
    def Position(self, Left=None, Top=None):
        try:
            if Left is not None:
                self._Left = Left
            if Top is not None:
                self._Top = Top
            if Left is not None or Top is not None:
                self._Frame.Position(Left=self._Left, Top=self._Top)
            return self._Frame.Position()
        except Exception as E:
            self._GUI.Error(f"{self._Type} -> Position -> {E}")
            
    def Size(self, Width=False, Height=False):
        try:
            if Width:
                self._Width = Width
            if Height:
                self._Height = Height
            if Width or Height:
                self._Frame.Size(Width=self._Width, Height=self._Height)
                self.Create()
            return self._Frame.Size()
        except Exception as E:
            self._GUI.Error(f"{self._Type} -> Size -> {E}")
    
    def Enlarge(self, Value=None):
        try:
            if Value is not None:
                self._Left -= Value
                self._Top -= Value
                self._Width += Value
                self._Height += Value
                self._Frame.Enlarge(Value)
                self.Create()
            return True
        except Exception as E:
            self._GUI.Error(f"{self._Type} -> Enlarge -> {E}")
    
    def Shrink(self, Value=None):
        try:
            if Value is not None:
                self._Left += Value
                self._Top += Value
                self._Width -= Value
                self._Height -= Value
                self._Frame.Shrink(Value)
                self.Create()
            return True
        except Exception as E:
            self._GUI.Error(f"{self._Type} -> Shrink -> {E}")
            
    def Box(self):
        try:
            return self._Frame.Box()
        except Exception as E:
            self._GUI.Error(f"{self._Type} -> Box -> {E}")
        
    def Ratio(self):
        try:
            return self._Frame.Ratio()
        except Exception as E:
            self._GUI.Error(f"{self._Type} -> Ratio -> {E}")
        
    def Locate(self, Width, Height, Left, Top):
        try:
            Width = self._Width*(Width/100)
            Height = self._Height*(Height/100)
            Left = self._Width*(Left/100)-self._Border_Size
            Top = self._Height*(Top/100)-self._Border_Size
            return [Width, Height, Left, Top]
        except Exception as E:
            self._GUI.Error(f"{self._Type} -> Locate -> {E}")
        
    def Locate_Reverse(self, Width, Height, Left, Top):
        try:
            Width = round((Width/self._Width)*100, 3)
            Height = round((Height/self._Height)*100, 3)
            Left =  round((Left/self._Width)*100, 3)
            Top =  round((Top/self._Height)*100, 3)
            return [Width, Height, Left, Top]
        except Exception as E:
            self._GUI.Error(f"{self._Type} -> Locate_Reverse -> {E}")
            
    def Create(self):
        try:
            if not self._Background:
                self._Background = self._Main._Background
                if not hasattr(self, "_Light_Background"):
                    setattr(self, "_Light_Background", self._Background)
                if not hasattr(self, "_Dark_Background"):
                    setattr(self, "_Dark_Background", self._GUI.Invert(self._Background))
            if self._Auto_Dark and not self._GUI._Dark_Mode:
                self.Update_Color()
            if not self._Initialized:
                self.Update_Color()
                self._Width_Current, self._Height_Current, self._Left_Current, self._Top_Current = self._Width, self._Height, self._Left, self._Top
                self._Frame.Config(Width=self._Width_Current, Height=self._Height_Current, Left=self._Left_Current, Top=self._Top_Current)
                self._Frame.Config(Background=self._Background, Border_Size=self._Border_Size, Border_Color=self._Border_Color)
                self._Frame.Create()
                Event_Bind(self._Widget, On_Hover_In=lambda E: self.On_Hover_In(E), On_Hover_Out=lambda E: self.On_Hover_Out(E))
                self._Font = TK.font.Font(family='Times New Roman', size=self._Font_Size, weight=self._Font_Weight)
                self._Widget.config(font=self._Font)
                self.Tag(Name='Default', Font_Size=self._Font_Size, Font_Weight=self._Font_Weight, Foreground=self._Foreground)
                if not self._Display:
                    self.Hide()
                self._Main._Widget.append(self)
                self._Initialized = True
            if self._Disable:
                State = TK.DISABLED
            else:
                State = TK.NORMAL
            self._Scrollbar_Vertical.config(orient="vertical", command=self._Widget.yview, width=self._Scrollbar_Size)
            self._Scrollbar_Horizontal.config(orient="horizontal", command=self._Widget.xview, width=self._Scrollbar_Size)
            self._Widget.config(background=self._Background, state=State, bd=0, highlightthickness=0, relief=TK.FLAT, wrap=self._Warp.lower())
            self._Widget.configure(yscrollcommand=self._Scrollbar_Vertical.set, xscrollcommand=self._Scrollbar_Horizontal.set)
            if self._Display:
                self.Show()
            else:
                self.Hide()
            if self._Name!=self._Last_Name:
                if self._Last_Name:
                    if self._Last_Name in self._Main.__dict__:
                        del self._Main.__dict__[self._Last_Name]
                if self._Name:
                    self._Main.__dict__[self._Name] = self
                self._Last_Name = self._Name
        except Exception as E:
            self._GUI.Error(f"{self._Type} -> Create -> {E}")
            
    def Update_Color(self):
        try:
            self._GUI.Initiate_Colors(self)
        except Exception as E:
            self._GUI.Error(f"{self._Type} -> Update_Color -> {E}")


import os
import threading, math, time
from io import BytesIO
import urllib
from PIL import Image as PIL_Image, ImageTk as PIL_ImageTk
from .N_Canvas_Image import Canvas_Image
from .N_Custom import Event_Bind_Canvas

class Canvas_Dynamic_Image(Canvas_Image):
    
    def __init__(self, Main):
        super().__init__(Main)
        self._Type = "Canvas_Dynamic_Image"
        self._Zoom_Scale = 1.0
        self._Zoom_Center = None
        self._Drag_Last_X = 0
        self._Drag_Last_Y = 0
        self._Image_Width = 0
        self._Image_Height = 0
        self._Hover_Active = False
        self._Canvas.Bind_Item(self._Widget, On_Click=self.Drag_Start)
        self._Canvas.Bind_Item(self._Widget, On_Drag=self.Drag)
        self._Canvas.Bind_Item(self._Widget, On_Right_Click=self.Reset)
        self._Canvas.Bind_Item(self._Widget, On_Hover_In=lambda E: self.Hover(True))
        self._Canvas.Bind_Item(self._Widget, On_Hover_Out=lambda E: self.Hover(False))
        self._Canvas.Bind(On_Mouse_Wheel=self.Zoom)

    def __str__(self):
        return "Nucleon_Glunoix_Canvas_Dynamic_Image[]"

    def __repr__(self):
        return "Nucleon_Glunoix_Canvas_Dynamic_Image[]"
            
    def Clear(self):
        try:
            super().Clear()
            self._Zoom_Scale = 1.0
            self._Zoom_Center = None
        except Exception as E:
            self._Canvas._GUI.Error(f"{self._Type} -> Clear -> {E}")

    def Open(self):
        try:
            super().Open()
            if self._Image:
                if not hasattr(self, "_Image_Width") or not hasattr(self, "_Image_Height") or not self._Image_Width or not self._Image_Height:
                    try:
                        self._Image_Width, self._Image_Height = self._Image.size
                    except Exception:
                        self._Image_Width, self._Image_Height = 0, 0
                if self._Image_Width and self._Image_Height:
                    if self._Zoom_Center is None:
                        self._Zoom_Center = (self._Image_Width // 2, self._Image_Height // 2)
                    CX, CY = self._Zoom_Center
                    CX = max(0, min(CX, self._Image_Width))
                    CY = max(0, min(CY, self._Image_Height))
                    self._Zoom_Center = (CX, CY)
                if not hasattr(self, "_Zoom_Scale") or self._Zoom_Scale < 1.0:
                    self._Zoom_Scale = 1.0
            else:
                self._Zoom_Scale = 1.0
                self._Zoom_Center = None
        except Exception as E:
            self._Canvas._GUI.Error(f"{self._Type} -> Open -> {E}")

    def Convert(self):
        try:
            if not self._Image:
                return False
            if self._Resize and self._Canvas._Type!='Scroll':
                Left, Right, Width, Height = self._Canvas.Box()
                Width_Ratio = Width/self._Canvas._Width if getattr(self._Canvas, "_Width", 0) else 1
                Height_Ratio = Height/self._Canvas._Height if getattr(self._Canvas, "_Height", 0) else 1
            else:
                Width_Ratio = 1
                Height_Ratio = 1
            Frame_Width = int(max(1, round(self._Width*Width_Ratio)))
            Frame_Height = int(max(1, round(self._Height*Height_Ratio)))
            if Frame_Width<=0 or Frame_Height<=0:
                return False
            if not getattr(self, "_Image_Width", 0) or not getattr(self, "_Image_Height", 0):
                try:
                    self._Image_Width, self._Image_Height = self._Image.size
                except Exception:
                    return False
            if not self._Zoom_Center:
                self._Zoom_Center = (self._Image_Width/2.0, self._Image_Height/2.0)
            Zoom_Width = self._Image_Width/self._Zoom_Scale
            Zoom_Height = self._Image_Height/self._Zoom_Scale
            Zoom_Width = max(1.0, min(Zoom_Width, float(self._Image_Width)))
            Zoom_Height = max(1.0, min(Zoom_Height, float(self._Image_Height)))
            CX, CY = self._Zoom_Center
            CX = max(Zoom_Width/2.0, min(self._Image_Width - Zoom_Width/2.0, CX))
            CY = max(Zoom_Height/2.0, min(self._Image_Height - Zoom_Height/2.0, CY))
            self._Zoom_Center = (CX, CY)
            Left_C = CX - Zoom_Width/2.0
            Top_C = CY - Zoom_Height/2.0
            Right_C = Left_C + Zoom_Width
            Bottom_C = Top_C + Zoom_Height
            Base_Image = self._Image
            Crop = Base_Image.crop((int(Left_C), int(Top_C), int(Right_C), int(Bottom_C)))
            Angle = getattr(self, "_Rotate", 0) + getattr(self, "_Angle", 0)
            if Angle:
                Crop = Crop.rotate(Angle, PIL_Image.NEAREST, expand=1)
            if self._Transparent:
                Crop = Crop.convert("RGBA")
                Mode = "RGBA"
                Background = (0, 0, 0, 0)
            else:
                Mode = Crop.mode if Crop.mode in ["RGB", "L"] else "RGB"
                Crop = Crop.convert(Mode)
                Background = 0 if Mode=="L" else (0, 0, 0)
            Image_Width, Image_Height = Crop.size
            if Image_Width<=0 or Image_Height<=0:
                return False
            Aspect_Ratio = getattr(self, "_Aspect_Ratio", True)
            Image_Aspect = Image_Width/Image_Height
            Frame_Aspect = Frame_Width/Frame_Height
            Use_Cover = Aspect_Ratio and self._Zoom_Scale>1.0001
            if not Aspect_Ratio:
                New_Width = Frame_Width
                New_Height = Frame_Height
                Crop = Crop.resize((New_Width, New_Height), PIL_Image.NEAREST)
                Frame_Image = Crop
            else:
                if Use_Cover:
                    if Image_Aspect>Frame_Aspect:
                        New_Height = Frame_Height
                        New_Width = int(max(1, round(Frame_Height*Image_Aspect)))
                    else:
                        New_Width = Frame_Width
                        New_Height = int(max(1, round(Frame_Width/Image_Aspect)))
                    if New_Width<=0 or New_Height<=0:
                        return False
                    Crop = Crop.resize((New_Width, New_Height), PIL_Image.NEAREST)
                    Offset_X = max(0, (New_Width - Frame_Width)//2)
                    Offset_Y = max(0, (New_Height - Frame_Height)//2)
                    Crop = Crop.crop((Offset_X, Offset_Y, Offset_X+Frame_Width, Offset_Y+Frame_Height))
                    Frame_Image = Crop
                else:
                    if Image_Aspect>Frame_Aspect:
                        New_Width = Frame_Width
                        New_Height = int(max(1, round(Frame_Width/Image_Aspect)))
                    else:
                        New_Height = Frame_Height
                        New_Width = int(max(1, round(Frame_Height*Image_Aspect)))
                    if New_Width<=0 or New_Height<=0:
                        return False
                    Crop = Crop.resize((New_Width, New_Height), PIL_Image.NEAREST)
                    Frame_Image = PIL_Image.new(Mode, (Frame_Width, Frame_Height), Background)
                    Offset_X = (Frame_Width - New_Width)//2
                    Offset_Y = (Frame_Height - New_Height)//2
                    if Mode=="RGBA":
                        Alpha = Crop.split()[3] if Crop.mode=="RGBA" else None
                        Frame_Image.paste(Crop, (Offset_X, Offset_Y), Alpha)
                    else:
                        Frame_Image.paste(Crop, (Offset_X, Offset_Y))
            W = Frame_Image.size[0]
            H = Frame_Image.size[1]
            Skew_X = max(-100.0, min(100.0, float(getattr(self, "_Skew_Horizontal", 0.0) or 0.0)))
            Skew_Y = max(-100.0, min(100.0, float(getattr(self, "_Skew_Vertical", 0.0) or 0.0)))
            Delta_X = min((W - 1) * 0.5, (abs(Skew_X) / 100.0) * (W * 0.5))
            Delta_Y = min((H - 1) * 0.5, (abs(Skew_Y) / 100.0) * (H * 0.5))
            TL_X = 0.0
            TL_Y = 0.0
            TR_X = float(W)
            TR_Y = 0.0
            BR_X = float(W)
            BR_Y = float(H)
            BL_X = 0.0
            BL_Y = float(H)
            if Skew_X > 0:
                TL_X = Delta_X
                TR_X = W - Delta_X
            elif Skew_X < 0:
                BL_X = Delta_X
                BR_X = W - Delta_X
            if Skew_Y > 0:
                TR_Y = Delta_Y
                BR_Y = H - Delta_Y
            elif Skew_Y < 0:
                TL_Y = Delta_Y
                BL_Y = H - Delta_Y
            if Delta_X == 0 and Delta_Y == 0:
                return PIL_ImageTk.PhotoImage(Frame_Image)
            Src_Points = [(0.0, 0.0), (float(W), 0.0), (float(W), float(H)), (0.0, float(H))]
            Dst_Points = [(TL_X, TL_Y), (TR_X, TR_Y), (BR_X, BR_Y), (BL_X, BL_Y)]
            A = []
            B_Vals = []
            for (X, Y), (U, V) in zip(Dst_Points, Src_Points):
                A.append([X, Y, 1, 0, 0, 0, -U * X, -U * Y])
                B_Vals.append(U)
                A.append([0, 0, 0, X, Y, 1, -V * X, -V * Y])
                B_Vals.append(V)
            N = 8
            for I in range(N):
                A[I].append(B_Vals[I])
            for Col in range(N):
                Pivot = max(range(Col, N), key=lambda R: abs(A[R][Col]))
                if abs(A[Pivot][Col]) < 1e-12:
                    continue
                if Pivot != Col:
                    A[Col], A[Pivot] = A[Pivot], A[Col]
                Pivot_Val = A[Col][Col]
                Inv_Pivot = 1.0 / Pivot_Val
                for J in range(Col, N + 1):
                    A[Col][J] *= Inv_Pivot
                for R in range(Col + 1, N):
                    Factor = A[R][Col]
                    if Factor != 0.0:
                        for J in range(Col, N + 1):
                            A[R][J] -= Factor * A[Col][J]
            Coeffs = [0.0] * N
            for I in reversed(range(N)):
                S = A[I][N]
                for J in range(I + 1, N):
                    S -= A[I][J] * Coeffs[J]
                Coeffs[I] = S
            try:
                Warped_Image = Frame_Image.transform((W, H), PIL_Image.PERSPECTIVE, Coeffs, resample=PIL_Image.BICUBIC, fillcolor=(0, 0, 0, 0) if self._Transparent and Mode=="RGBA" else None)
            except TypeError:
                Background_Image = PIL_Image.new(Mode, (W, H), Background)
                Warped_Image = Frame_Image.transform((W, H), PIL_Image.PERSPECTIVE, Coeffs, resample=PIL_Image.BICUBIC)
                if self._Transparent and Warped_Image.mode != "RGBA":
                    Warped_Image = Warped_Image.convert("RGBA")
                if Mode=="RGBA":
                    Background_Image.paste(Warped_Image, (0, 0), Warped_Image.split()[3])
                else:
                    Background_Image.paste(Warped_Image, (0, 0))
                Warped_Image = Background_Image
            return PIL_ImageTk.PhotoImage(Warped_Image)
        except Exception as E:
            self._Canvas._GUI.Error(f"{self._Type} -> Convert -> {E}")
            return False

    def Load(self):
        try:
            if not self._Image:
                return
            if getattr(self, "_Photo", False):
                Image = self._Image
            else:
                Image = self.Convert()
            if not Image:
                return
            self._Canvas._Frame.itemconfig(self._Widget, image=Image)
            self._TK_Image = Image
        except Exception as E:
            self._Canvas._GUI.Error(f"{self._Type} -> Load -> {E}")
                
    def Drag_Start(self, Event):
        try:
            self._Drag_Last_X = self._Canvas._Frame.canvasx(Event.x)
            self._Drag_Last_Y = self._Canvas._Frame.canvasy(Event.y)
        except Exception:
            self._Drag_Last_X, self._Drag_Last_Y = 0, 0

    def Drag(self, Event):
        try:
            if not self._Image or not self._Zoom_Center:
                return
            Curr_X = self._Canvas._Frame.canvasx(Event.x)
            Curr_Y = self._Canvas._Frame.canvasy(Event.y)
            Delta_X = Curr_X - self._Drag_Last_X
            Delta_Y = Curr_Y - self._Drag_Last_Y
            self._Drag_Last_X = Curr_X
            self._Drag_Last_Y = Curr_Y
            View_Width = self._Image_Width/self._Zoom_Scale
            View_Height = self._Image_Height/self._Zoom_Scale
            Box = self._Canvas._Frame.bbox(self._Widget)
            if not Box:
                return
            X1, Y1, X2, Y2 = Box
            Inner_Width = max(1, X2 - X1)
            Inner_Height = max(1, Y2 - Y1)
            Move_X_Image = -Delta_X*(View_Width/Inner_Width)
            Move_Y_Image = -Delta_Y*(View_Height/Inner_Height)
            New_Center_X = self._Zoom_Center[0] + Move_X_Image
            New_Center_Y = self._Zoom_Center[1] + Move_Y_Image
            Half_Width = View_Width/2.0
            Half_Height = View_Height/2.0
            New_Center_X = max(Half_Width, min(self._Image_Width - Half_Width, New_Center_X))
            New_Center_Y = max(Half_Height, min(self._Image_Height - Half_Height, New_Center_Y))
            self._Zoom_Center = (New_Center_X, New_Center_Y)
            self.Load()
        except Exception as E:
            self._Canvas._GUI.Error(f"{self._Type} -> Drag -> {E}")

    def Hover(self, Value):
        try:
            self._Hover_Active = Value
        except Exception as E:
            self._Canvas._GUI.Error(f"{self._Type} -> Hover -> {E}")

    def Zoom(self, Event):
        try:
            if not self._Hover_Active:
                return
            if not self._Image:
                return
            Zoom_Factor = 1.1 if getattr(Event, "delta", 0)>0 else 0.9
            New_Zoom_Scale = self._Zoom_Scale*Zoom_Factor
            New_Zoom_Scale = max(1.0, New_Zoom_Scale)
            if New_Zoom_Scale == self._Zoom_Scale:
                return
            Canvas_X = self._Canvas._Frame.canvasx(Event.x)
            Canvas_Y = self._Canvas._Frame.canvasy(Event.y)
            Box = self._Canvas._Frame.bbox(self._Widget)
            if not Box:
                return
            X1, Y1, X2, Y2 = Box
            Inner_Width = max(1, X2 - X1)
            Inner_Height = max(1, Y2 - Y1)
            Mouse_Rel_X = (Canvas_X - X1)/Inner_Width
            Mouse_Rel_Y = (Canvas_Y - Y1)/Inner_Height
            View_Width = self._Image_Width/self._Zoom_Scale
            View_Height = self._Image_Height/self._Zoom_Scale
            View_Left = self._Zoom_Center[0] - View_Width/2.0
            View_Top = self._Zoom_Center[1] - View_Height/2.0
            Mouse_Abs_X = View_Left + Mouse_Rel_X*View_Width
            Mouse_Abs_Y = View_Top + Mouse_Rel_Y*View_Height
            New_View_Width = self._Image_Width/New_Zoom_Scale
            New_View_Height = self._Image_Height/New_Zoom_Scale
            New_View_Left = Mouse_Abs_X - Mouse_Rel_X*New_View_Width
            New_View_Top = Mouse_Abs_Y - Mouse_Rel_Y*New_View_Height
            New_Center_X = New_View_Left + New_View_Width/2.0
            New_Center_Y = New_View_Top + New_View_Height/2.0
            Half_Width = New_View_Width/2.0
            Half_Height = New_View_Height/2.0
            New_Center_X = max(Half_Width, min(self._Image_Width - Half_Width, New_Center_X))
            New_Center_Y = max(Half_Height, min(self._Image_Height - Half_Height, New_Center_Y))
            self._Zoom_Scale = New_Zoom_Scale
            self._Zoom_Center = (New_Center_X, New_Center_Y)
            self.Load()
        except Exception as E:
            self._Canvas._GUI.Error(f"{self._Type} -> Zoom -> {E}")

    def Reset(self, Event=False):
        try:
            self._Angle = 0
            self._Zoom_Scale = 1.0
            if self._Image and getattr(self, "_Image_Width", 0) and getattr(self, "_Image_Height", 0):
                self._Zoom_Center = (self._Image_Width//2, self._Image_Height//2)
            self.Load()
        except Exception as E:
            self._Canvas._GUI.Error(f"{self._Type} -> Reset -> {E}")



# IMPORT LIBRARIES
import threading, math, time
from PIL import Image as PIL_Image, ImageDraw as PIL_ImageDraw, ImageFont as PIL_ImageFont

class Canvas_Text:
    
    def __init__(self, Main):
        self._Canvas = Main
        self._Main = Main
        self._Config = ['Name', 'Width', 'Height', 'Left', 'Top', 'Color', 'Background', 'Size', 'Value', 'Weight', 'Font', 'Resize', 'Justify', 'Anchor', 'Resize_Font', 'Vertical', 'Rotate', 'Skew_Horizontal', 'Skew_Vertical']
        self._Display = True
        self._Resize = True
        self._Name = False
        self._Last_Name = False
        self._Resize_Font = True
        self._Type = 'Canvas_Text'
        self._Left, self._Top, self._Width, self._Height = 0, 0, 0, 0
        self._Value = ''
        self._Color = '#000000'
        self._Background = False
        self._Size = 20
        self._Weight = 'normal'
        self._Font = 'Helvetica'
        self._Anchor = 'nw'
        self._Justify = 'center'
        self._Vertical = False
        self._Rotate = 0
        self._Skew_Horizontal = 0
        self._Skew_Vertical = 0
        self._Widget = self._Canvas.Image()
        self._Widget.Config(Name = False, Transparent = True)
        self._Canvas._Item.append(self._Widget)

    def __str__(self):
        return "Nucleon_Glunoix_Canvas_Text[]"

    def __repr__(self):
        return "Nucleon_Glunoix_Canvas_Text[]"

    def Copy(self, Name=False, Main=False):
        try:
            if not Main:
                Main = self._Canvas
            if Main._Type in ['Canvas', 'Scroll', 'Group']:
                Temp_Main = Main
                Temp_Type = Temp_Main._Type
                while Temp_Type=='Group':
                    Temp_Main = Temp_Main._Main
                    Temp_Type = Temp_Main._Type
                if Temp_Type=='Canvas' or Temp_Type=='Scroll':
                    Instance = type(self)(Main)
                    for Key in self._Config:
                        if hasattr(self, "_"+Key):
                            setattr(Instance, "_"+Key, getattr(self, "_"+Key))
                    if Name:
                        setattr(Instance, "_Name", Name)
                    Instance.Create()
                    return Instance
                else:
                    raise Exception('Widget can only copy to Canvas/Scroll')
            else:
                raise Exception('Widget can only copy to Canvas/Scroll')
        except Exception as E:
            self._Canvas._GUI.Error(f"{self._Type} -> Copy -> {E}")

    def Delete(self):
        try:
            self._Widget.Delete()
            if self in self._Canvas._Widget:
                self._Canvas._Widget.remove(self)
            if self:
                del self
        except Exception as Error:
            self._Canvas._GUI.Error(f"{self._Type} -> Delete -> {Error}")

    def Hide(self):
        try:
            self._Widget.Hide()
            self._Display = False
        except Exception as Error:
            self._Canvas._GUI.Error(f"{self._Type} -> Hide -> {Error}")

    def Show(self):
        try:
            self._Widget.Show()
            self._Display = True
        except Exception as Error:
            self._Canvas._GUI.Error(f"{self._Type} -> Show -> {Error}")

    def Display(self):
        try:
            self._Widget.Display()
            self._Display = True
        except Exception as Error:
            self._Canvas._GUI.Error(f"{self._Type} -> Display -> {Error}")

    def Set(self, Value):
        try:
            self._Value = Value
            self.Create()
        except Exception as Error:
            self._Canvas._GUI.Error(f"{self._Type} -> Set -> {Error}")
            
    def Animate(self):
        try:
            self._Widget.Animate()
            self.Show()
        except Exception as Error:
            self._Canvas._GUI.Error(f"{self._Type} -> Animate -> {Error}")
            self.Animate_Cancel()
            
    def Animate_Cancel(self):
        try:
            self._Widget.Animate_Cancel()
        except Exception as Error:
            self._Canvas._GUI.Error(f"{self._Type} -> Animate_Cancel -> {Error}")
            
    def Config_Get(self, *Input):
        try:
            Return = {}
            for Each in self._Config:
                if Each in Input:
                    Return[Each] = getattr(self, "_" + Each)
            return Return
        except Exception as Error:
            self._Canvas._GUI.Error(f"{self._Type} -> Config_Get -> {Error}")

    def Config(self, **Input):
        try:
            Run = False
            for Each in self._Config:
                if Each in Input:
                    Value = Input[Each]
                    setattr(self, "_" + Each, Value)
                    setattr(self, "_" + Each + "_Current", Value)
                    Run = True
            Forward = {Key: Val for Key, Val in Input.items() if Key not in ('Name', 'Transparent')}
            if Forward:
                self._Widget.Config(**Forward)
            if Run:
                self.Create()
        except Exception as Error:
            self._Canvas._GUI.Error(f"{self._Type} -> Config -> {Error}")

    def Move(self, Left = None, Top = None):
        try:
            if Left is not None:
                self._Left += Left
            if Top is not None:
                self._Top += Top
            if Left is not None or Top is not None:
                self.Position(Left = self._Left, Top = self._Top)
            return True
        except Exception as Error:
            self._Canvas._GUI.Error(f"{self._Type} -> Move -> {Error}")
            
    def Center(self, Left = None, Top = None):
        try:
            if Left is not None:
                self._Left = Left - self._Width / 2
            if Top is not None:
                self._Top = Top - self._Height / 2
            if Left is not None or Top is not None:
                self.Position(Left = self._Left, Top = self._Top)
            return [self._Left + self._Width / 2, self._Top + self._Height / 2]
        except Exception as Error:
            self._Canvas._GUI.Error(f"{self._Type} -> Center -> {Error}")
        
    def Position(self, Left = None, Top = None):
        try:
            if Left is not None:
                self._Left = Left
            if Top is not None:
                self._Top = Top
            if Left is not None or Top is not None:
                self._Widget.Position(self._Left, self._Top)
                self.Create()
            return [self._Left, self._Top]
        except Exception as Error:
            self._Canvas._GUI.Error(f"{self._Type} -> Position -> {Error}")
            
    def Size(self, Width = False, Height = False):
        try:
            if Width:
                self._Width = Width
            if Height:
                self._Height = Height
            if Width or Height:
                self._Widget.Size(self._Width, self._Height)
                self.Create()
            return [self._Width, self._Height]
        except Exception as Error:
            self._Canvas._GUI.Error(f"{self._Type} -> Size -> {Error}")
    
    def Enlarge(self, Value=None):
        try:
            if Value is not None:
                self._Left -= Value
                self._Top -= Value
                self._Width += Value
                self._Height += Value
                self._Widget.Enlarge(Value)
                self.Create()
            return True
        except Exception as E:
            self._Canvas._GUI.Error(f"{self._Type} -> Enlarge -> {E}")
    
    def Shrink(self, Value=None):
        try:
            if Value is not None:
                self._Left += Value
                self._Top += Value
                self._Width -= Value
                self._Height -= Value
                self._Widget.Shrink(Value)
                self.Create()
            return True
        except Exception as E:
            self._Canvas._GUI.Error(f"{self._Type} -> Shrink -> {E}")
            
    def Box(self):
        try:
            return self._Widget.Box()
        except Exception as E:
            self._Canvas._GUI.Error(f"{self._Type} -> Box -> {E}")
        
    def Ratio(self):
        try:
            return self._Widget.Ratio()
        except Exception as E:
            self._GUI.Error(f"{self._Type} -> Ratio -> {E}")

    def Bind(self, **Input):
        try:
            self._Widget.Bind(**Input)
        except Exception as Error:
            self._Canvas._GUI.Error(f"{self._Type} -> Bind -> {Error}")

    def Create(self):
        try:
            Image_Object = self.Render()
            self._Widget._Pil = True
            self._Widget.Set(Image_Object)
            if self._Name!=self._Last_Name:
                if self._Last_Name:
                    if self._Last_Name in self._Canvas.__dict__:
                        del self._Canvas.__dict__[self._Last_Name]
                if self._Name:
                    self._Canvas.__dict__[self._Name] = self
                self._Last_Name = self._Name
        except Exception as Error:
            self._Canvas._GUI.Error(f"{self._Type} -> Create -> {Error}")
            
    def Resize(self, Event):
        try:
            self._Canvas._Frame.itemconfigure(self._Widget, state='normal')
            self._Canvas._Frame.tag_raise(self._Widget)
            if self._Display:
                self.Create()
        except Exception as E:
            self._Canvas._GUI.Error(f"{self._Type} -> Resize -> {E}")

    def Render(self):
        try:
            Text_Value = self._Value if isinstance(self._Value, str) else str(self._Value)
            if self._Vertical:
                S = Text_Value.replace('\t', '    ')
                S = ''.join('\u00A0' if ch == ' ' else ch for ch in S)
                Text_Value = '\n'.join(list(S))
            Font_Size = int(max(1, self._Size_Current if hasattr(self, '_Size_Current') else self._Size))
            try:
                Font_Object = PIL_ImageFont.truetype(self._Font.lower() + '.ttf', Font_Size)
            except:
                Font_Object = PIL_ImageFont.truetype('arial.ttf', Font_Size)
            Weight_Value = (self._Weight or 'normal').strip().lower()
            Is_Bold = Weight_Value in ('bold', 'b', '700', 'semibold', 'demibold', 'heavy', 'black')
            if Is_Bold:
                if Font_Size <= 30:
                    Stroke_Width = 0.5
                elif Font_Size <= 55:
                    Stroke_Width = 1
                else:
                    Stroke_Width = 2
            else:
                Stroke_Width = 0
            Width_Target = int(self._Width if self._Width else 1)
            Height_Target = int(self._Height if self._Height else 1)
            Temp_Image = PIL_Image.new('RGBA', (1, 1), (0, 0, 0, 0))
            Draw = PIL_ImageDraw.Draw(Temp_Image)
            def Measure_Width(S):
                B = Draw.textbbox((0, 0), S, font=Font_Object, stroke_width=Stroke_Width)
                return max(1, B[2] - B[0])
            Paragraphs = Text_Value.split('\n') if '\n' in Text_Value else [Text_Value]
            Lines = []
            for Paragraph in Paragraphs:
                if not self._Width:
                    Lines.append(Paragraph)
                    continue
                Words = Paragraph.split(' ')
                if len(Words) == 1:
                    Current = ''
                    for Char in Paragraph:
                        Test = Current + Char
                        if Measure_Width(Test) <= Width_Target or not Current:
                            Current = Test
                        else:
                            Lines.append(Current)
                            Current = Char
                    if Current:
                        Lines.append(Current)
                    continue
                Current = ''
                for Word in Words:
                    Test = Current + (' ' if Current else '') + Word
                    if Measure_Width(Test) <= Width_Target or not Current:
                        Current = Test
                    else:
                        Lines.append(Current)
                        Current = Word
                if Current:
                    Lines.append(Current)
            if not Lines:
                Lines = ['']
            Para_Text = '\n'.join(Lines)
            Align_For_Measure, Pos_X, Pos_Y = self.Justify(self._Justify, Width_Target, Height_Target, Para_Text, Font_Object, Stroke_Width)
            B_All = Draw.multiline_textbbox((0, 0), Para_Text, font=Font_Object, align=Align_For_Measure, spacing=0, stroke_width=Stroke_Width)
            Text_Width = max(1, B_All[2] - B_All[0])
            Text_Height = max(1, B_All[3] - B_All[1])
            if not self._Width:
                Width_Target = Text_Width
            if not self._Height:
                Height_Target = Text_Height
            Background_Color = self._Background if self._Background else (0, 0, 0, 0)
            Image_Object = PIL_Image.new('RGBA', (max(1, Width_Target), max(1, Height_Target)), Background_Color)
            Draw = PIL_ImageDraw.Draw(Image_Object)
            Fill_Color = self._Color if self._Color else (0, 0, 0, 0)
            Align, Pos_X, Pos_Y = self.Justify(self._Justify, Width_Target, Height_Target, Para_Text, Font_Object, Stroke_Width)
            Draw.multiline_text((Pos_X, Pos_Y), Para_Text, fill=Fill_Color, font=Font_Object, align=Align, spacing=0, stroke_width=Stroke_Width, stroke_fill=Fill_Color)
            self._Size_Rendered = int(max(1, self._Size_Current if hasattr(self, '_Size_Current') else self._Size))
            return Image_Object
        except Exception as Error:
            self._Canvas._GUI.Error(f"{self._Type} -> Render -> {Error}")
            return PIL_Image.new('RGBA', (max(1, int(self._Width or 1)), max(1, int(self._Height or 1))), (0, 0, 0, 0))

    def Justify(self, Justify_Value, Width_Target, Height_Target, Para_Text, Font_Object, Stroke_Width):
        try:
            J = (Justify_Value or 'left').strip().lower()
            Temp = PIL_Image.new('RGBA', (1, 1), (0, 0, 0, 0))
            D = PIL_ImageDraw.Draw(Temp)
            def Bbox_For(Align_Mode):
                B = D.multiline_textbbox((0, 0), Para_Text, font=Font_Object, align=Align_Mode, spacing=0, stroke_width=Stroke_Width)
                return B, max(1, B[2] - B[0]), max(1, B[3] - B[1])
            if J in ('left', 'l'):
                B, Tw, Th = Bbox_For('left')
                return 'left', -B[0], (Height_Target - Th) / 2 - B[1]
            if J in ('center', 'centre', 'c'):
                B, Tw, Th = Bbox_For('center')
                return 'center', (Width_Target - Tw) / 2 - B[0], (Height_Target - Th) / 2 - B[1]
            if J in ('right', 'r'):
                B, Tw, Th = Bbox_For('right')
                return 'right', Width_Target - Tw - B[0], (Height_Target - Th) / 2 - B[1]
            B, Tw, Th = Bbox_For('left')
            if J == 'n':
                return 'left', (Width_Target - Tw) / 2 - B[0], -B[1]
            if J == 's':
                return 'left', (Width_Target - Tw) / 2 - B[0], Height_Target - Th - B[1]
            if J == 'w':
                return 'left', -B[0], (Height_Target - Th) / 2 - B[1]
            if J == 'e':
                return 'left', Width_Target - Tw - B[0], (Height_Target - Th) / 2 - B[1]
            if J == 'nw':
                return 'left', -B[0], -B[1]
            if J == 'ne':
                return 'left', Width_Target - Tw - B[0], -B[1]
            if J == 'sw':
                return 'left', -B[0], Height_Target - Th - B[1]
            if J == 'se':
                return 'left', Width_Target - Tw - B[0], Height_Target - Th - B[1]
            return 'left', -B[0], (Height_Target - Th) / 2 - B[1]
        except Exception as Error:
            self._Canvas._GUI.Error(f"{self._Type} -> Justify -> {Error}")
            return 'left', 0, 0